#ifndef VEDITAREA_H
#define VEDITAREA_H

#include <QWidget>
#include <QJsonObject>
#include <QString>
#include <QFileInfo>
#include <QDir>
#include <QVector>
#include <QPair>
#include <QtDebug>
#include <QSplitter>
#include "vnotebook.h"
#include "veditwindow.h"
#include "vtoc.h"
#include "vnavigationmode.h"

class VNote;
class VFile;
class VDirectory;
class VFindReplaceDialog;
class QLabel;
class VVim;

class VEditArea : public QWidget, public VNavigationMode
{
    Q_OBJECT
public:
    explicit VEditArea(VNote *vnote, QWidget *parent = 0);
    bool isFileOpened(const VFile *p_file);
    bool closeAllFiles(bool p_forced);
    bool closeFile(const VFile *p_file, bool p_forced);
    bool closeFile(const VDirectory *p_dir, bool p_forced);
    bool closeFile(const VNotebook *p_notebook, bool p_forced);
    // Returns current edit tab.
    VEditTab *currentEditTab();
    // Returns the count of VEditWindow.
    inline int windowCount() const;
    // Returns the index of @p_window.
    int windowIndex(const VEditWindow *p_window) const;
    // Move tab widget @p_widget from window @p_fromIdx to @p_toIdx.
    // @p_widget has been removed from the original window.
    // If fail, just delete the p_widget.
    void moveTab(QWidget *p_widget, int p_fromIdx, int p_toIdx);
    inline VFindReplaceDialog *getFindReplaceDialog() const;
    // Return selected text of current edit tab.
    QString getSelectedText();
    void splitCurrentWindow();
    void removeCurrentWindow();
    // Focus next window (curWindowIndex + p_biaIdx).
    // Return the new current window index, otherwise, return -1.
    int focusNextWindow(int p_biaIdx);
    void moveCurrentTabOneSplit(bool p_right);
    VEditWindow *getCurrentWindow() const;

    // Implementations for VNavigationMode.
    void registerNavigation(QChar p_majorKey) Q_DECL_OVERRIDE;
    void showNavigation() Q_DECL_OVERRIDE;
    void hideNavigation() Q_DECL_OVERRIDE;
    bool handleKeyNavigation(int p_key, bool &p_succeed) Q_DECL_OVERRIDE;

signals:
    // Emit when current window's tab status updated.
    void tabStatusUpdated(const VEditTabInfo &p_info);

    void outlineChanged(const VToc &toc);
    void curHeaderChanged(const VAnchor &anchor);

    // Emit when want to show message in status bar.
    void statusMessage(const QString &p_msg);

    // Emit when Vim status updated.
    void vimStatusUpdated(const VVim *p_vim);

protected:
    void mousePressEvent(QMouseEvent *event) Q_DECL_OVERRIDE;

public slots:
    void openFile(VFile *p_file, OpenFileMode p_mode);
    void editFile();
    void saveFile();
    void readFile();
    void saveAndReadFile();
    void handleOutlineItemActivated(const VAnchor &anchor);
    void handleFileUpdated(const VFile *p_file);
    void handleDirectoryUpdated(const VDirectory *p_dir);
    void handleNotebookUpdated(const VNotebook *p_notebook);

private slots:
    // Split @curWindow via inserting a new window around it.
    // @p_right: insert the new window on the right side.
    void splitWindow(VEditWindow *p_window, bool p_right = true);

    void handleRemoveSplitRequest(VEditWindow *curWindow);
    void handleWindowFocused();
    void handleOutlineChanged(const VToc &toc);
    void handleCurHeaderChanged(const VAnchor &anchor);
    void handleFindTextChanged(const QString &p_text, uint p_options);
    void handleFindOptionChanged(uint p_options);
    void handleFindNext(const QString &p_text, uint p_options, bool p_forward);
    void handleReplace(const QString &p_text, uint p_options,
                       const QString &p_replaceText, bool p_findNext);
    void handleReplaceAll(const QString &p_text, uint p_options,
                          const QString &p_replaceText);
    void handleFindDialogClosed();

    // Hanle the status update of current tab within a VEditWindow.
    void handleWindowTabStatusUpdated(const VEditTabInfo &p_info);

    // Handle the statusMessage signal of VEditWindow.
    void handleWindowStatusMessage(const QString &p_msg);

    // Handle the vimStatusUpdated signal of VEditWindow.
    void handleWindowVimStatusUpdated(const VVim *p_vim);

private:
    void setupUI();
    QVector<QPair<int, int> > findTabsByFile(const VFile *p_file);
    int openFileInWindow(int windowIndex, VFile *p_file, OpenFileMode p_mode);
    void setCurrentTab(int windowIndex, int tabIndex, bool setFocus);
    void setCurrentWindow(int windowIndex, bool setFocus);
    inline VEditWindow *getWindow(int windowIndex) const;
    void insertSplitWindow(int idx);
    void removeSplitWindow(VEditWindow *win);

    // Update status of current window.
    void updateWindowStatus();

    VNote *vnote;
    int curWindowIndex;

    // Splitter holding multiple split windows
    QSplitter *splitter;
    VFindReplaceDialog *m_findReplace;

    // Navigation Mode.
    // Map second key to VEditWindow.
    QMap<QChar, VEditWindow *> m_keyMap;
    QVector<QLabel *> m_naviLabels;
};

inline VEditWindow* VEditArea::getWindow(int windowIndex) const
{
    Q_ASSERT(windowIndex < splitter->count());
    return dynamic_cast<VEditWindow *>(splitter->widget(windowIndex));
}

inline int VEditArea::windowCount() const
{
    return splitter->count();
}

inline VFindReplaceDialog *VEditArea::getFindReplaceDialog() const
{
    return m_findReplace;
}

#endif // VEDITAREA_H
